@@ -46,6 +46,7 @@ void ngtcp2_strm_init(ngtcp2_strm *strm, int64_t stream_id, uint32_t flags,
   strm->tx.streamfrq = NULL;
   strm->tx.offset = 0;
   strm->tx.max_offset = max_tx_offset;
+  strm->tx.last_blocked_offset = UINT64_MAX;
   strm->tx.last_max_stream_data_ts = UINT64_MAX;
   strm->tx.loss_count = 0;
   strm->tx.last_lost_pkt_num = -1;
@@ -376,18 +377,16 @@ int ngtcp2_strm_streamfrq_pop(ngtcp2_strm *strm, ngtcp2_frame_chain **pfrc,
   fr = &frc->fr.stream;
   datalen = ngtcp2_vec_len(fr->data, fr->datacnt);
 
-  if (left == 0) {
-    /* datalen could be zero if 0 length STREAM has been sent */
-    if (datalen || ngtcp2_ksl_len(strm->tx.streamfrq) > 1) {
-      rv = ngtcp2_ksl_insert(strm->tx.streamfrq, NULL, &fr->offset, frc);
-      if (rv != 0) {
-        assert(ngtcp2_err_is_fatal(rv));
-        ngtcp2_frame_chain_objalloc_del(frc, strm->frc_objalloc, strm->mem);
-        return rv;
-      }
-      *pfrc = NULL;
-      return 0;
+  /* datalen could be zero if 0 length STREAM has been sent */
+  if (left == 0 && datalen) {
+    rv = ngtcp2_ksl_insert(strm->tx.streamfrq, NULL, &fr->offset, frc);
+    if (rv != 0) {
+      assert(ngtcp2_err_is_fatal(rv));
+      ngtcp2_frame_chain_objalloc_del(frc, strm->frc_objalloc, strm->mem);
+      return rv;
     }
+    *pfrc = NULL;
+    return 0;
   }
 
   if (datalen > left) {
