@@ -1,4 +1,4 @@
-/*  Copyright (C) 2022 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>
+/*  Copyright (C) 2023 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -70,6 +70,10 @@ int check_ref_dflt(
 	knotd_conf_check_args_t *args
 );
 
+int check_ref_empty(
+	knotd_conf_check_args_t *args
+);
+
 int check_listen(
 	knotd_conf_check_args_t *args
 );
@@ -78,6 +82,14 @@ int check_xdp_listen(
 	knotd_conf_check_args_t *args
 );
 
+int check_cert_pin(
+	knotd_conf_check_args_t *args
+);
+
+int check_modulo(
+	knotd_conf_check_args_t *args
+);
+
 int check_database(
 	knotd_conf_check_args_t *args
 );
@@ -90,10 +102,6 @@ int check_module_id(
 	knotd_conf_check_args_t *args
 );
 
-int check_file(
-	knotd_conf_check_args_t *args
-);
-
 int check_server(
 	knotd_conf_check_args_t *args
 );
@@ -126,6 +134,10 @@ int check_remotes(
 	knotd_conf_check_args_t *args
 );
 
+int check_dnskey_sync(
+	knotd_conf_check_args_t *args
+);
+
 int check_catalog_group(
 	knotd_conf_check_args_t *args
 );
