@@ -156,6 +156,9 @@ static void policy_load(knot_kasp_policy_t *policy, conf_t *conf, conf_val_t *id
 	}
 	policy->ds_push = conf_val_count(&val) > 0;
 
+	val = conf_id_get(conf, C_POLICY, C_DNSKEY_SYNC, id);
+	policy->has_dnskey_sync = (val.code == KNOT_EOK);
+
 	val = conf_id_get(conf, C_POLICY, C_OFFLINE_KSK, id);
 	policy->offline_ksk = conf_bool(&val);
 
