@@ -1,4 +1,4 @@
-/*  Copyright (C) 2021 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>
+/*  Copyright (C) 2023 CZ.NIC, z.s.p.o. <knot-dns@labs.nic.cz>
 
     This program is free software: you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
@@ -107,8 +107,8 @@ static knotd_state_t dnsproxy_fwd(knotd_state_t state, knot_pkt_t *pkt,
 	}
 	const struct sockaddr_storage *dst = &proxy->remote;
 	const struct sockaddr_storage *src = &proxy->via;
-	knot_request_t *req = knot_request_make(re.mm, dst, src, qdata->query, NULL,
-	                                        flags);
+	knot_request_t *req = knot_request_make_generic(re.mm, dst, src, qdata->query,
+	                                                NULL, NULL, NULL, NULL, 0, flags);
 	if (req == NULL) {
 		knot_requestor_clear(&re);
 		return state; /* Ignore, not enough memory. */
