@@ -60,6 +60,8 @@ typedef enum {
 
 	SEM_ERR_DS_RDATA_ALG,
 	SEM_ERR_DS_RDATA_DIGLEN,
+	SEM_ERR_DS_APEX,
+	SEM_ERR_DS_NONDELEG,
 
 	SEM_ERR_DNSKEY_NONE,
 	SEM_ERR_DNSKEY_INVALID,
