@@ -21,6 +21,7 @@
 #include "knot/catalog/catalog_db.h"
 #include "utils/common/msg.h"
 #include "utils/common/params.h"
+#include "utils/common/signal.h"
 #include "utils/common/util_conf.h"
 
 #define PROGRAM_NAME	"kcatalogprint"
@@ -28,6 +29,8 @@
 static knot_dname_t *filter_member = NULL;
 static knot_dname_t *filter_catalog = NULL;
 
+signal_ctx_t signal_ctx = { 0 }; // global, needed by signal handler
+
 static void print_help(void)
 {
 	printf("Usage: %s [-c | -C | -D <path>] [parameters]\n"
@@ -95,6 +98,8 @@ static void params_cleanup(void)
 
 int main(int argc, char *argv[])
 {
+	catalog_t c = { { 0 } };
+
 	struct option opts[] = {
 		{ "config",  required_argument, NULL, 'c' },
 		{ "confdb",  required_argument, NULL, 'C' },
@@ -106,6 +111,9 @@ int main(int argc, char *argv[])
 		{ NULL }
 	};
 
+	signal_ctx.close_db = &c.db;
+	signal_init_std();
+
 	int opt = 0;
 	while ((opt = getopt_long(argc, argv, "c:C:D:a:m:hV", opts, NULL)) != -1) {
 		switch (opt) {
@@ -159,8 +167,6 @@ int main(int argc, char *argv[])
 		goto failure;
 	}
 
-	catalog_t c = { { 0 } };
-
 	char *db = conf_db(conf(), C_CATALOG_DB);
 	catalog_init(&c, db, 0); // mapsize grows automatically
 	free(db);
